import numpy as np
import pytest

from autolens.data.imaging import image
from autolens.data.array import mask as mask
from autolens.model.inversion import inversions, pixelizations, regularization
from autolens.lensing import lensing_fitting
from autolens.model.galaxy import galaxy as g
from autolens.data.fitting import fitting
from autolens.data.fitting.util import fitting_util
from autolens.lensing import lensing_image
from autolens.lensing import plane as pl
from autolens.lensing import ray_tracing
from autolens.model.profiles import light_profiles as lp
from test.mock.mock_profiles import MockLightProfile
from test.mock.mock_galaxy import MockHyperGalaxy
from test.mock.mock_lensing import MockTracer


@pytest.fixture(name="no_galaxies", scope='function')
def make_no_galaxies():
    return [g.Galaxy()]


@pytest.fixture(name="sersic")
def make_sersic():
    return lp.EllipticalSersic(axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6, sersic_index=4.0)


@pytest.fixture(name="galaxy_light", scope='function')
def make_galaxy_light(sersic):
    return g.Galaxy(light_profile=sersic)


@pytest.fixture(name='li_no_blur')
def make_li_no_blur():
    im = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)
    im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    ma = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    ma = mask.Mask(array=ma, pixel_scale=1.0)

    return lensing_image.LensingImage(im, ma, sub_grid_size=2)


@pytest.fixture(name='li_blur')
def make_li_blur():
    im = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0]])), pixel_scale=1.0, renormalize=False)
    im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    ma = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    ma = mask.Mask(array=ma, pixel_scale=1.0)

    return lensing_image.LensingImage(im, ma, sub_grid_size=2)


@pytest.fixture(name='li_no_blur_1x1')
def make_li_no_blur_1x1():
    psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)

    im = image.Image(np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

    ma = mask.Mask(array=np.array([[True, True, True],
                                   [True, False, True],
                                   [True, True, True]]), pixel_scale=1.0)

    return lensing_image.LensingImage(im, ma, sub_grid_size=1)


@pytest.fixture(name='li_manual')
def make_li_manual():
    im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 6.0, 0.0],
                   [0.0, 7.0, 8.0, 9.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 5.0, 1.0],
                                     [3.0, 4.0, 0.0]])), pixel_scale=1.0)
    im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
    ma = mask.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=1.0)

    return lensing_image.LensingImage(im, ma, sub_grid_size=1)

@pytest.fixture(name='li_manual_1')
def make_li_manual_1():
    im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 6.0, 0.0],
                   [0.0, 7.0, 8.0, 10.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 1.0, 1.0],
                                     [3.0, 2.0, 0.0]])), pixel_scale=1.0)
    im = image.Image(im, pixel_scale=2.0, psf=psf, noise_map=2.0 * np.ones((5, 5)))
    ma = mask.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=2.0)

    return lensing_image.LensingImage(im, ma, sub_grid_size=2)


@pytest.fixture(name='hyper')
def make_hyper():
    class Hyper(object):

        def __init__(self):
            pass

    hyper = Hyper()

    hyper.hyper_model_image = np.array([1.0, 3.0, 5.0, 7.0, 9.0, 8.0, 6.0, 4.0, 0.0])
    hyper.hyper_galaxy_images = [np.array([1.0, 3.0, 5.0, 7.0, 9.0, 8.0, 6.0, 4.0, 0.0]),
                                 np.array([1.0, 3.0, 5.0, 7.0, 9.0, 8.0, 6.0, 4.0, 0.0])]
    hyper.hyper_minimum_values = [0.2, 0.8]
    return hyper

@pytest.fixture(name='li_hyper_no_blur')
def make_li_hyper_no_blur(hyper):
    im = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)
    im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    ma = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    ma = mask.Mask(array=ma, pixel_scale=1.0)

    return lensing_image.LensingHyperImage(im, ma, hyper_model_image=hyper.hyper_model_image,
                                           hyper_galaxy_images=hyper.hyper_galaxy_images,
                                           hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=1)

@pytest.fixture(name='li_hyper_no_blur_1x1')
def make_li_hyper_no_blur_1x1(hyper):
    psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)

    im = image.Image(np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

    ma = mask.Mask(array=np.array([[True, True, True],
                                   [True, False, True],
                                   [True, True, True]]), pixel_scale=1.0)

    return lensing_image.LensingHyperImage(im, ma, hyper_model_image=hyper.hyper_model_image,
                                           hyper_galaxy_images=hyper.hyper_galaxy_images,
                                           hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=1)

@pytest.fixture(name='li_hyper_manual')
def make_li_hyper_manual(hyper):
    im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 6.0, 0.0],
                   [0.0, 7.0, 8.0, 9.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 5.0, 1.0],
                                     [3.0, 4.0, 0.0]])), pixel_scale=1.0)
    im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
    ma = mask.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=1.0)

    return lensing_image.LensingHyperImage(im, ma, hyper_model_image=hyper.hyper_model_image,
                                           hyper_galaxy_images=hyper.hyper_galaxy_images,
                                           hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=1)

@pytest.fixture(name='li_hyper_manual_1')
def make_li_hyper_manual_1(hyper):
    im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 9.0, 0.0],
                   [0.0, 7.0, 8.0, 9.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = image.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 5.0, 1.0],
                                     [3.0, 4.0, 0.0]])), pixel_scale=2.0)
    im = image.Image(im, pixel_scale=2.0, psf=psf, noise_map=np.ones((5, 5)))
    ma = mask.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=1.0)

    return lensing_image.LensingHyperImage(im, ma, hyper_model_image=hyper.hyper_model_image,
                                           hyper_galaxy_images=hyper.hyper_galaxy_images,
                                           hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=2)


class TestUnmaskedModelImages:

    def test___of_galaxies__x1_galaxy__3x3_padded_image__no_psf_blurring(self, galaxy_light):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.padded_grids])

        manual_model_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(tracer.image_plane_images_[0])
        manual_model_image_0 = psf.convolve(manual_model_image_0)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
            lensing_image=li, tracer=tracer, image_index=0)

        assert (manual_model_image_0[1:4, 1:4] == padded_model_images[0][0]).all()

    def test___of_galaxies__x1_galaxy__3x3_padded_image__asymetric_psf_blurring(self, galaxy_light):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.padded_grids])

        manual_model_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(tracer.image_plane_images_[0])
        manual_model_image_0 = psf.convolve(manual_model_image_0)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
            lensing_image=li, tracer=tracer, image_index=0)

        assert (manual_model_image_0[1:4, 1:4] == padded_model_images[0][0]).all()

    def test___of_galaxies__x2_galaxies__3x3_padded_image__asymetric_psf_blurring(self):
        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1], image_plane_grids=[li.padded_grids])

        manual_model_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][0])
        manual_model_image_0 = psf.convolve(manual_model_image_0)

        manual_model_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][1])
        manual_model_image_1 = psf.convolve(manual_model_image_1)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
            lensing_image=li, tracer=tracer, image_index=0)

        assert (manual_model_image_0[1:4, 1:4] == padded_model_images[0][0]).all()
        assert (manual_model_image_1[1:4, 1:4] == padded_model_images[0][1]).all()

    def test___same_as_above_but_image_and_souce_plane(self):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grids=[li.padded_grids])

        manual_model_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][0])
        manual_model_image_0 = psf.convolve(manual_model_image_0)

        manual_model_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][1])
        manual_model_image_1 = psf.convolve(manual_model_image_1)


        manual_model_image_2 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[0][0])
        manual_model_image_2 = psf.convolve(manual_model_image_2)

        manual_model_image_3 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[0][1])
        manual_model_image_3 = psf.convolve(manual_model_image_3)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
            lensing_image=li, tracer=tracer, image_index=0)

        assert (manual_model_image_0[1:4, 1:4] == padded_model_images[0][0]).all()
        assert (manual_model_image_1[1:4, 1:4] == padded_model_images[0][1]).all()
        assert (manual_model_image_2[1:4, 1:4] == padded_model_images[1][0]).all()
        assert (manual_model_image_3[1:4, 1:4] == padded_model_images[1][1]).all()

    def test___same_as_above_but_x2_images(self):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)

        im_0 = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))
        li_0 = lensing_image.LensingImage(im_0, ma, sub_grid_size=1)

        im_1 = image.Image(array=2.0 * np.ones((3, 3)), pixel_scale=2.0, psf=psf, noise_map=np.ones((3, 3)))
        li_1 = lensing_image.LensingImage(im_1, ma, sub_grid_size=2)

        g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.1), centre=(0.1, 0.1))
        g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.2))
        g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.3))
        g3 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=0.4))

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g2, g3],
                                                     image_plane_grids=[li_0.padded_grids, li_1.padded_grids])

        manual_model_image_0_of_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][0])
        manual_model_image_0_of_image_0 = psf.convolve(manual_model_image_0_of_image_0)

        manual_model_image_1_of_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[0][1])
        manual_model_image_1_of_image_0 = psf.convolve(manual_model_image_1_of_image_0)

        manual_model_image_2_of_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[0][0])
        manual_model_image_2_of_image_0 = psf.convolve(manual_model_image_2_of_image_0)

        manual_model_image_3_of_image_0 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[0][1])
        manual_model_image_3_of_image_0 = psf.convolve(manual_model_image_3_of_image_0)

        manual_model_image_0_of_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[1][0])
        manual_model_image_0_of_image_1 = psf.convolve(manual_model_image_0_of_image_1)

        manual_model_image_1_of_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.image_plane.image_plane_images_of_galaxies_[1][1])
        manual_model_image_1_of_image_1 = psf.convolve(manual_model_image_1_of_image_1)

        manual_model_image_2_of_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[1][0])
        manual_model_image_2_of_image_1 = psf.convolve(manual_model_image_2_of_image_1)

        manual_model_image_3_of_image_1 = tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(
            tracer.source_plane.image_plane_images_of_galaxies_[1][1])
        manual_model_image_3_of_image_1 = psf.convolve(manual_model_image_3_of_image_1)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_images_and_tracer(
            lensing_images=[li_0, li_1], tracer=tracer)

        assert (manual_model_image_0_of_image_0[1:4, 1:4] == padded_model_images[0][0][0]).all()
        assert (manual_model_image_1_of_image_0[1:4, 1:4] == padded_model_images[0][0][1]).all()
        assert (manual_model_image_2_of_image_0[1:4, 1:4] == padded_model_images[0][1][0]).all()
        assert (manual_model_image_3_of_image_0[1:4, 1:4] == padded_model_images[0][1][1]).all()

        assert (manual_model_image_0_of_image_1[1:4, 1:4] == padded_model_images[1][0][0]).all()
        assert (manual_model_image_1_of_image_1[1:4, 1:4] == padded_model_images[1][0][1]).all()
        assert (manual_model_image_2_of_image_1[1:4, 1:4] == padded_model_images[1][1][0]).all()
        assert (manual_model_image_3_of_image_1[1:4, 1:4] == padded_model_images[1][1][1]).all()

    def test__properties_of_fit(self, galaxy_light):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.padded_grids])

        manual_model_image_0 = \
            padded_tracer.image_plane.grids[0].regular.map_to_2d_keep_padded(padded_tracer.image_plane_images_[0])
        manual_model_image_0 = psf.convolve(manual_model_image_0)

        padded_model_images = lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
            lensing_image=li, tracer=padded_tracer, image_index=0)

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.grids])

        fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer, padded_tracer=padded_tracer)

        assert (manual_model_image_0[1:4, 1:4] == fit.unmasked_model_profile_images[0]).all()
        assert (padded_model_images[0][0] == fit.unmasked_model_profile_images_of_galaxies[0][0][0]).all()

    def test__padded_tracer_is_none__returns_none(self, galaxy_light):

        psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                         [0.0, 1.0, 0.0],
                                         [0.0, 0.0, 0.0]])), pixel_scale=1.0)
        im = image.Image(array=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

        ma = mask.Mask(array=np.array([[True, True, True],
                                       [True, False, True],
                                       [True, True, True]]), pixel_scale=1.0)
        li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.grids])

        fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer, padded_tracer=None)

        assert fit.unmasked_model_profile_images == None
        assert fit.unmasked_model_profile_images_of_galaxies == None


class TestLensingProfileFit:

    class TestModelImages:

        def test__mock_tracer__2x2_image_all_1s__3x3_psf_all_1s__blurring_region__image_blurs_to_9s(self, li_blur):

            tracer = MockTracer(images=[li_blur.mask.map_2d_array_to_masked_1d_array(li_blur.image)],
                                blurring_images=[np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])],
                                has_light_profile=True, has_hyper_galaxy=False, has_pixelization=False)

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li_blur], tracer=tracer)

            assert (fit.model_datas[0] == np.array([[0.0, 0.0, 0.0, 0.0],
                                                 [0.0, 9.0, 9.0, 0.0],
                                                 [0.0, 9.0, 9.0, 0.0],
                                                 [0.0, 0.0, 0.0, 0.0]])).all()

        def test__real_tracer__2x2_image__no_psf_blurring(self, li_no_blur, galaxy_light):

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li_no_blur.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li_no_blur], tracer=tracer)

            tracer_image = tracer.image_plane_images_[0]
            assert (tracer_image == fit.model_datas_[0]).all()

            tracer_image_2d = li_no_blur.grids.regular.scaled_array_from_array_1d(tracer_image)
            assert (tracer_image_2d == fit.model_datas[0]).all()

        def test__real_tracer__2x2_image__psf_is_non_symmetric_producing_l_shape(self, galaxy_light):
            psf = image.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                             [0.0, 2.0, 1.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)
            im = image.Image(array=np.ones((4, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

            ma = mask.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)
            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            # Manually compute result of convolution, which is each central value *2.0 plus its 2 appropriate neighbors

            central_values = tracer.image_plane_images_[0]
            blurring_values = tracer.image_plane_blurring_images_[0]

            tracer_blurred_image_manual_0 = 2.0 * central_values[0] + 3.0 * central_values[2] + blurring_values[4]
            tracer_blurred_image_manual_1 = 2.0 * central_values[1] + 3.0 * central_values[3] + central_values[0]
            tracer_blurred_image_manual_2 = 2.0 * central_values[2] + 3.0 * blurring_values[9] + blurring_values[6]
            tracer_blurred_image_manual_3 = 2.0 * central_values[3] + 3.0 * blurring_values[10] + central_values[2]

            assert tracer_blurred_image_manual_0 == pytest.approx(fit.model_datas_[0][0], 1e-6)
            assert tracer_blurred_image_manual_1 == pytest.approx(fit.model_datas_[0][1], 1e-6)
            assert tracer_blurred_image_manual_2 == pytest.approx(fit.model_datas_[0][2], 1e-6)
            assert tracer_blurred_image_manual_3 == pytest.approx(fit.model_datas_[0][3], 1e-6)

            assert tracer_blurred_image_manual_0 == pytest.approx(fit.model_datas[0][1, 1], 1e-6)
            assert tracer_blurred_image_manual_1 == pytest.approx(fit.model_datas[0][1, 2], 1e-6)
            assert tracer_blurred_image_manual_2 == pytest.approx(fit.model_datas[0][2, 1], 1e-6)
            assert tracer_blurred_image_manual_3 == pytest.approx(fit.model_datas[0][2, 2], 1e-6)

        def test__model_images_of_planes__real_tracer__2x2_image__psf_is_non_symmetric_producing_l_shape(self):

            psf = image.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                             [0.0, 2.0, 1.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)
            im = image.Image(array=np.ones((4, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

            ma = mask.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)
            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g1],
                                                         image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            model_lens_image = li.convolver_image.convolve_image(tracer.image_plane.image_plane_images_[0],
                                                                 tracer.image_plane.image_plane_blurring_images_[0])

            model_source_image = li.convolver_image.convolve_image(tracer.source_plane.image_plane_images_[0],
                                                                   tracer.source_plane.image_plane_blurring_images_[0])

            model_lens_image = li.grids.regular.scaled_array_from_array_1d(model_lens_image)
            model_source_image = li.grids.regular.scaled_array_from_array_1d(model_source_image)

            assert (fit.model_images_of_planes[0][0] == model_lens_image).all()
            assert (fit.model_images_of_planes[0][1] == model_source_image).all()

        def test__same_as_above_but_multi_tracer(self):
            
            psf = image.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                             [0.0, 2.0, 1.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)
            im = image.Image(array=np.ones((4, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

            ma = mask.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)
            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), redshift=0.1)
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0), redshift=0.2)
            g2 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=3.0), redshift=0.3)

            from astropy import cosmology as cosmo

            tracer = ray_tracing.TracerMulti(galaxies=[g0, g1, g2], image_plane_grids=[li.grids],
                                             cosmology=cosmo.Planck15)

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            model_plane_image_0 = li.convolver_image.convolve_image(tracer.planes[0].image_plane_images_[0],
                                                                    tracer.planes[0].image_plane_blurring_images_[0])

            model_plane_image_1 = li.convolver_image.convolve_image(tracer.planes[1].image_plane_images_[0],
                                                                    tracer.planes[1].image_plane_blurring_images_[0])
            
            model_plane_image_2 = li.convolver_image.convolve_image(tracer.planes[2].image_plane_images_[0],
                                                                    tracer.planes[2].image_plane_blurring_images_[0])

            model_plane_image_0 = li.grids.regular.scaled_array_from_array_1d(model_plane_image_0)
            model_plane_image_1 = li.grids.regular.scaled_array_from_array_1d(model_plane_image_1)
            model_plane_image_2 = li.grids.regular.scaled_array_from_array_1d(model_plane_image_2)

            assert (fit.model_images_of_planes[0][0] == model_plane_image_0).all()
            assert (fit.model_images_of_planes[0][1] == model_plane_image_1).all()
            assert (fit.model_images_of_planes[0][2] == model_plane_image_2).all()

        def test__model_images_of_planes__if_galaxy_has_no_light_profile__replace_with_none(self):
            psf = image.PSF(array=(np.array([[0.0, 3.0, 0.0],
                                             [0.0, 2.0, 1.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)
            im = image.Image(array=np.ones((4, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))
            ma = mask.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)
            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g0_image = pl.intensities_from_grid(grid=li.grids.sub, galaxies=[g0])
            g0_blurring_image = pl.intensities_from_grid(grid=li.grids.blurring, galaxies=[g0])
            g0_model_image = li.grids.regular.scaled_array_from_array_1d(li.convolver_image.convolve_image(g0_image, g0_blurring_image))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            assert (fit.model_images_of_planes[0][0] == g0_model_image).all()
            assert (fit.model_images_of_planes[0][1] == g0_model_image).all()

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g.Galaxy()],
                                                         image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            assert (fit.model_images_of_planes[0][0] == g0_model_image).all()
            assert fit.model_images_of_planes[0][1] == None

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g0],
                                                         image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            assert fit.model_images_of_planes[0][0] == None
            assert (fit.model_images_of_planes[0][1] == g0_model_image).all()

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g.Galaxy()],
                                                         image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            assert fit.model_images_of_planes[0][0] == None
            assert fit.model_images_of_planes[0][1] == None

    class TestLikelihood:

        def test__1x1_image__tracing_fits_data_perfectly__no_psf_blurring__lh_is_noise_term(self):
            psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            im = image.Image(np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

            ma = mask.Mask(array=np.array([[True, True, True],
                                           [True, False, True],
                                           [True, True, True]]), pixel_scale=1.0)
            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__tracing_fits_data_with_chi_sq_5(self):

            psf = image.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            im = image.Image(5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4)))
            im[1,2]  = 4.0

            ma = mask.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)

            li = lensing_image.LensingImage(im, ma, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0, size=2))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grids=[li.grids])

            fit = lensing_fitting.LensingProfileFit(lensing_images=[li], tracer=tracer)

            assert fit.chi_squared_term == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.likelihood == -0.5 * (25.0 + 2.0*np.log(2 * np.pi * 1.0))

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grids=[li.grids, li.grids])
            fit = lensing_fitting.LensingProfileFit(lensing_images=[li, li], tracer=tracer)

            assert fit.chi_squared_terms[0] == 25.0
            assert fit.reduced_chi_squareds[0] == 25.0 / 2.0
            assert fit.likelihoods[0] == -0.5 * (25.0 + 2.0*np.log(2 * np.pi * 1.0))

            assert fit.chi_squared_terms[1] == 25.0
            assert fit.reduced_chi_squareds[1] == 25.0 / 2.0
            assert fit.likelihoods[1] == -0.5 * (25.0 + 2.0*np.log(2 * np.pi * 1.0))

            assert fit.chi_squared_term == 50.0
            assert fit.reduced_chi_squared == 50.0 / 2.0
            assert fit.likelihood == -0.5 * (50.0 + 4.0*np.log(2 * np.pi * 1.0))

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_manual):

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grids=[li_manual.grids])

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                                  image_plane_grids=[li_manual.padded_grids])

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_manual, tracer=tracer,
                                                                padded_tracer=padded_tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im, blurring_images_=blurring_im,
                                                                          convolvers=[li_manual.convolver_image])
            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_manual], model_datas_=model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_manual.noise_map_])

            assert li_manual.grids.regular.scaled_array_from_array_1d(li_manual.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) == \
                   pytest.approx(fit.model_data, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_manual.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms,
                                                                                noise_terms=noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(lensing_image=li_manual,
                                                                                            tracer=tracer)
            assert fast_likelihood == pytest.approx(fit.likelihood)

            padded_model_image = fitting_util.unmasked_blurred_images_from_fitting_images(fitting_images=[li_manual],
                                                                                     unmasked_images_=padded_tracer.image_plane_images_)

            assert (padded_model_image == fit.unmasked_model_profile_image).all()

            padded_model_image_of_galaxies = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image=li_manual,
                                                                                                tracer=padded_tracer,
                                                                                                image_index=0)

            assert (padded_model_image_of_galaxies[0][0] == fit.unmasked_model_profile_images_of_galaxies[0][0][0]).all()
            assert (padded_model_image_of_galaxies[1][0] == fit.unmasked_model_profile_images_of_galaxies[0][1][0]).all()

        def test___manual_image_and_psf__x2_image(self, li_manual, li_manual_1):

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grids=[li_manual.grids, li_manual_1.grids])

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                                  image_plane_grids=[li_manual.padded_grids,
                                                                                     li_manual_1.padded_grids])

            fit = lensing_fitting.fit_multiple_lensing_images_with_tracer(lensing_images=[li_manual, li_manual_1],
                                                                          tracer=tracer, padded_tracer=padded_tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im, blurring_images_=blurring_im,
                                                                          convolvers=[li_manual.convolver_image,
                                                                                     li_manual_1.convolver_image])
            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_manual, li_manual_1],
                                                                     model_datas_=model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_manual.noise_map_,
                                                                                           li_manual_1.noise_map_])

            assert li_manual.grids.regular.scaled_array_from_array_1d(li_manual.noise_map_) == \
                   pytest.approx(fit.noise_maps[0], 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) == \
                   pytest.approx(fit.model_datas[0], 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residuals[0], 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squareds[0], 1e-4)

            assert li_manual_1.grids.regular.scaled_array_from_array_1d(li_manual_1.noise_map_) == \
                   pytest.approx(fit.noise_maps[1], 1e-4)
            assert li_manual_1.grids.regular.scaled_array_from_array_1d(model_images[1]) == \
                   pytest.approx(fit.model_datas[1], 1e-4)
            assert li_manual_1.grids.regular.scaled_array_from_array_1d(residuals[1]) == \
                   pytest.approx(fit.residuals[1], 1e-4)
            assert li_manual_1.grids.regular.scaled_array_from_array_1d(chi_squareds[1]) == \
                   pytest.approx(fit.chi_squareds[1], 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_manual.noise_map_, li_manual_1.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms,
                                                                                noise_terms=noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihoods[0], 1e-4)
            assert likelihoods[1] == pytest.approx(fit.likelihoods[1], 1e-4)
            assert likelihoods[0] + likelihoods[1] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = lensing_fitting.fast_likelihood_from_multiple_lensing_images_and_tracer(
                lensing_images=[li_manual, li_manual_1], tracer=tracer)
            assert fast_likelihood == pytest.approx(fit.likelihood)

            padded_model_images = fitting_util.unmasked_blurred_images_from_fitting_images(fitting_images=[li_manual, li_manual_1],
                                                                                      unmasked_images_=padded_tracer.image_plane_images_)

            assert (padded_model_images[0] == fit.unmasked_model_profile_images[0]).all()
            assert (padded_model_images[1] == fit.unmasked_model_profile_images[1]).all()

            padded_model_image_of_galaxies_0 = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image=li_manual,
                                                                                                tracer=padded_tracer,
                                                                                                image_index=0)

            assert (padded_model_image_of_galaxies_0[0][0] == fit.unmasked_model_profile_images_of_galaxies[0][0][0]).all()
            assert (padded_model_image_of_galaxies_0[1][0] == fit.unmasked_model_profile_images_of_galaxies[0][1][0]).all()

            padded_model_image_of_galaxies_1 = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image=li_manual_1,
                                                                                                tracer=padded_tracer,
                                                                                                image_index=1)

            assert (padded_model_image_of_galaxies_1[0][0] == fit.unmasked_model_profile_images_of_galaxies[1][0][0]).all()
            assert (padded_model_image_of_galaxies_1[1][0] == fit.unmasked_model_profile_images_of_galaxies[1][1][0]).all()


class TestHyperLensingProfileFit:

    class TestScaledLikelihood:

        def test__hyper_galaxy_adds_to_noise_term_for_scaled_noise__chi_squared_is_0(self, li_hyper_no_blur_1x1):
            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            g1 = g.Galaxy(light_profile=MockLightProfile(value=0.0))

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1],
                                                  image_plane_grids=[li_hyper_no_blur_1x1.grids])

            li_hyper_no_blur_1x1.hyper_model_image = np.array([1.0])
            li_hyper_no_blur_1x1.hyper_galaxy_images = [np.array([1.0]), np.array([1.0])]

            tracer.image_plane.galaxies[0].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=1.0,
                                                                          noise_power=1.0)
            tracer.image_plane.galaxies[1].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=2.0,
                                                                          noise_power=1.0)

            fit = lensing_fitting.HyperLensingProfileFit(lensing_hyper_images=[li_hyper_no_blur_1x1], tracer=tracer)

            chi_squared_term = 0.0
            scaled_noise_term = np.log(2 * np.pi * 4.0 ** 2.0)

            assert fit.scaled_likelihood == -0.5 * (chi_squared_term + scaled_noise_term)

        def test__hyper_galaxy_adds_to_noise_term_for_scaled_noise__chi_squared_nonzero(self, no_galaxies,
                                                                                        li_hyper_no_blur_1x1):
            li_hyper_no_blur_1x1[0] = 2.0

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            g1 = g.Galaxy(light_profile=MockLightProfile(value=0.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=no_galaxies,
                                                         image_plane_grids=[li_hyper_no_blur_1x1.grids])

            li_hyper_no_blur_1x1.hyper_model_image = np.array([1.0])
            li_hyper_no_blur_1x1.hyper_galaxy_images = [np.array([1.0]), np.array([1.0])]

            tracer.image_plane.galaxies[0].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=1.0,
                                                                          noise_power=1.0)
            tracer.image_plane.galaxies[1].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=2.0,
                                                                          noise_power=1.0)

            fit = lensing_fitting.HyperLensingProfileFit(lensing_hyper_images=[li_hyper_no_blur_1x1], tracer=tracer)

            scaled_chi_squared_term = (1.0 / (4.0)) ** 2.0
            scaled_noise_term = np.log(2 * np.pi * 4.0 ** 2.0)

            assert fit.scaled_likelihood == -0.5 * (scaled_chi_squared_term + scaled_noise_term)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=no_galaxies,
                                                         image_plane_grids=[li_hyper_no_blur_1x1.grids,
                                                                            li_hyper_no_blur_1x1.grids])

            tracer.image_plane.galaxies[0].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=1.0,
                                                                          noise_power=1.0)
            tracer.image_plane.galaxies[1].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=2.0,
                                                                          noise_power=1.0)

            fit = lensing_fitting.HyperLensingProfileFit(lensing_hyper_images=[li_hyper_no_blur_1x1,
                                                                               li_hyper_no_blur_1x1], tracer=tracer)

            scaled_chi_squared_term = (1.0 / (4.0)) ** 2.0
            scaled_noise_term = np.log(2 * np.pi * 4.0 ** 2.0)

            assert fit.scaled_chi_squared_terms[0] == scaled_chi_squared_term
            assert fit.scaled_chi_squared_terms[1] == scaled_chi_squared_term
            assert fit.scaled_noise_terms[0] == scaled_noise_term
            assert fit.scaled_noise_terms[1] == scaled_noise_term

            assert fit.scaled_likelihoods[0] == -0.5 * (scaled_chi_squared_term + scaled_noise_term)
            assert fit.scaled_likelihoods[1] == -0.5 * (scaled_chi_squared_term + scaled_noise_term)
            assert fit.scaled_likelihood == 2.0 * -0.5 * (scaled_chi_squared_term + scaled_noise_term)

    class TestAbstractLogic:

        def test__logic_in_abstract_fit(self, li_hyper_no_blur, galaxy_light):
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light],
                                                  image_plane_grids=[li_hyper_no_blur.grids])

            fit = lensing_fitting.HyperLensingProfileFit(lensing_hyper_images=[li_hyper_no_blur], tracer=tracer)

            assert fit.total_inversions == 0

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_hyper_manual):

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grids=[li_hyper_manual.grids])

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                                image_plane_grids=[li_hyper_manual.padded_grids])

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_hyper_manual, tracer=tracer,
                                                                padded_tracer=padded_tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im, blurring_images_=blurring_im,
                                                                          convolvers=[li_hyper_manual.convolver_image])
            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_hyper_manual], model_datas_=model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_hyper_manual.noise_map_])

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(li_hyper_manual.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) == \
                   pytest.approx(fit.model_image, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=li_hyper_manual.noise_map_)
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms,
                                                                                noise_terms=noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihood, 1e-4)

            contributions = fitting_util.contributions_from_hyper_images_and_galaxies(li_hyper_manual.hyper_model_image,
                                                                                 li_hyper_manual.hyper_galaxy_images, [hyper_galaxy, hyper_galaxy],
                                                                                 li_hyper_manual.hyper_minimum_values)

            scaled_noise_maps = fitting_util.scaled_noises_from_fitting_hyper_images_contributions_and_hyper_galaxies(
                fitting_hyper_images=[li_hyper_manual], contributions_=[contributions], hyper_galaxies=[hyper_galaxy,
                                                                                                        hyper_galaxy])
            scaled_chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, scaled_noise_maps)

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[0]) == \
                   pytest.approx(fit.contributions[0][0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[1]) == \
                   pytest.approx(fit.contributions[0][1], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_noise_maps[0]) == \
                   pytest.approx(fit.scaled_noise_map, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_chi_squareds[0]) == \
                   pytest.approx(fit.scaled_chi_squared, 1e-4)

            scaled_chi_squared_terms= fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=scaled_chi_squareds)
            scaled_noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=scaled_noise_maps)
            scaled_likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(
                chi_squared_terms=scaled_chi_squared_terms, noise_terms=scaled_noise_terms)

            assert scaled_likelihoods[0] == pytest.approx(fit.scaled_likelihood, 1e-4)

            fast_scaled_likelihood = lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(
                lensing_image=li_hyper_manual, tracer=tracer)

            assert fast_scaled_likelihood == fit.scaled_likelihood

            padded_model_image = fitting_util.unmasked_blurred_images_from_fitting_images(fitting_images=[li_hyper_manual],
                                                                                     unmasked_images_=padded_tracer.image_plane_images_)

            assert (padded_model_image == fit.unmasked_model_profile_images[0]).all()

            padded_model_image_of_galaxies = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(
                    lensing_image=li_hyper_manual, tracer=padded_tracer, image_index=0)

            assert (padded_model_image_of_galaxies[0][0] == fit.unmasked_model_profile_images_of_galaxies[0][0]).all()

        def test___same_as_above__x2_images(self, li_hyper_manual, li_hyper_manual_1):

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grids=[li_hyper_manual.grids, 
                                                                            li_hyper_manual_1.grids])

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                                image_plane_grids=[li_hyper_manual.padded_grids, 
                                                                                   li_hyper_manual_1.padded_grids])

            fit = lensing_fitting.fit_multiple_lensing_images_with_tracer(lensing_images=[li_hyper_manual,
                  li_hyper_manual_1], tracer=tracer,
                                                                          padded_tracer=padded_tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im, blurring_images_=blurring_im,
                                                                          convolvers=[li_hyper_manual.convolver_image,
                                                                                     li_hyper_manual_1.convolver_image])
            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_hyper_manual, li_hyper_manual_1],
                                                                     model_datas_=model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_hyper_manual.noise_map_,
                                                                                           li_hyper_manual_1.noise_map_])

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(li_hyper_manual.noise_map_) == \
                   pytest.approx(fit.noise_maps[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) == \
                   pytest.approx(fit.model_datas[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residuals[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squareds[0], 1e-4)

            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(li_hyper_manual_1.noise_map_) == \
                   pytest.approx(fit.noise_maps[1], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(model_images[1]) == \
                   pytest.approx(fit.model_datas[1], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(residuals[1]) == \
                   pytest.approx(fit.residuals[1], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(chi_squareds[1]) == \
                   pytest.approx(fit.chi_squareds[1], 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_hyper_manual.noise_map_,
                                                                           li_hyper_manual_1.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms,
                                                                                noise_terms=noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihoods[0], 1e-4)
            assert likelihoods[1] == pytest.approx(fit.likelihoods[1], 1e-4)
            assert likelihoods[0] + likelihoods[1] == pytest.approx(fit.likelihood, 1e-4)

            contributions = fitting_util.contributions_from_fitting_hyper_images_and_hyper_galaxies(
                fitting_hyper_images=[li_hyper_manual, li_hyper_manual_1],
                hyper_galaxies=[hyper_galaxy, hyper_galaxy])

            scaled_noise_maps = fitting_util.scaled_noises_from_fitting_hyper_images_contributions_and_hyper_galaxies(
                fitting_hyper_images=[li_hyper_manual, li_hyper_manual_1], 
                contributions_=contributions, hyper_galaxies=[hyper_galaxy, hyper_galaxy])

            scaled_chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, scaled_noise_maps)

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[0][0]) == \
                   pytest.approx(fit.contributions[0][0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[0][1]) == \
                   pytest.approx(fit.contributions[0][1], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_noise_maps[0]) == \
                   pytest.approx(fit.scaled_noise_maps[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_chi_squareds[0]) == \
                   pytest.approx(fit.scaled_chi_squareds[0], 1e-4)
            
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(contributions[1][0]) == \
                   pytest.approx(fit.contributions[1][0], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(contributions[1][1]) == \
                   pytest.approx(fit.contributions[1][1], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(scaled_noise_maps[1]) == \
                   pytest.approx(fit.scaled_noise_maps[1], 1e-4)
            assert li_hyper_manual_1.grids.regular.scaled_array_from_array_1d(scaled_chi_squareds[1]) == \
                   pytest.approx(fit.scaled_chi_squareds[1], 1e-4)

            scaled_chi_squared_terms= fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=scaled_chi_squareds)
            scaled_noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=scaled_noise_maps)
            scaled_likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(
                chi_squared_terms=scaled_chi_squared_terms, noise_terms=scaled_noise_terms)

            assert scaled_likelihoods[0] == pytest.approx(fit.scaled_likelihoods[0], 1e-4)
            assert scaled_likelihoods[1] == pytest.approx(fit.scaled_likelihoods[1], 1e-4)
            assert scaled_likelihoods[0] + scaled_likelihoods[1] == pytest.approx(fit.scaled_likelihood, 1e-4)

            fast_scaled_likelihood = lensing_fitting.fast_likelihood_from_multiple_lensing_images_and_tracer(
                lensing_images=[li_hyper_manual, li_hyper_manual_1], tracer=tracer)

            assert fast_scaled_likelihood == pytest.approx(fit.scaled_likelihood, 1e-2)

            padded_model_images = fitting_util.unmasked_blurred_images_from_fitting_images(
                fitting_images=[li_hyper_manual, li_hyper_manual_1], unmasked_images_=padded_tracer.image_plane_images_)

            assert (padded_model_images[0] == fit.unmasked_model_profile_images[0]).all()
            assert (padded_model_images[1] == fit.unmasked_model_profile_images[1]).all()

            padded_model_image_of_galaxies_0 = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image=li_hyper_manual,
                                                                                                tracer=padded_tracer,
                                                                                                image_index=0)

            assert (padded_model_image_of_galaxies_0[0][0] == fit.unmasked_model_profile_images_of_galaxies[0][0][0]).all()
            assert (padded_model_image_of_galaxies_0[1][0] == fit.unmasked_model_profile_images_of_galaxies[0][1][0]).all()

            padded_model_image_of_galaxies_1 = \
                lensing_fitting.unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image=li_hyper_manual_1,
                                                                                                tracer=padded_tracer,
                                                                                                image_index=1)

            assert (padded_model_image_of_galaxies_1[0][0] == fit.unmasked_model_profile_images_of_galaxies[1][0][0]).all()
            assert (padded_model_image_of_galaxies_1[1][0] == fit.unmasked_model_profile_images_of_galaxies[1][1][0]).all()


class TestInversionLensingFit:

    class TestAbstractLogic:

        def test__logic_in_abstract_fit(self, li_no_blur):

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li_no_blur.grids],
                                                         border=None)

            fit = lensing_fitting.LensingInversionFit(lensing_images=[li_no_blur], tracer=tracer)

            assert fit.total_inversions == 1

    class TestModelImageOfPlanes:

        def test__model_images_are_none_and_an_image(self):

            im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 0.0, 0.0, 0.0, 0.0]]).view(image.Image)

            ma = np.array([[True, True, True, True, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, True, True, True, True]])

            ma = mask.Mask(ma, pixel_scale=1.0)

            psf = image.PSF(array=np.array([[0.0, 0.0, 0.0],
                                            [0.0, 1.0, 0.0],
                                            [0.0, 0.0, 0.0]]), pixel_scale=1.0)
            im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
            li = lensing_image.LensingImage(im, ma, sub_grid_size=2)

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)))
            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li.grids], border=None)
            fit = lensing_fitting.LensingInversionFit(lensing_images=[li], tracer=tracer)

            assert fit.model_images_of_planes[0][0] == None
            assert fit.model_images_of_planes[0][1] == pytest.approx(np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), 1e-2)

    class TestRectangularInversion:

        def test__image_all_1s__direct_image_to_source_mapping__perfect_fit_even_with_regularization(self):
            im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 0.0, 0.0, 0.0, 0.0]]).view(image.Image)

            ma = np.array([[True, True, True, True, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, True, True, True, True]])
            ma = mask.Mask(ma, pixel_scale=1.0)

            psf = image.PSF(array=np.array([[0.0, 0.0, 0.0],
                                            [0.0, 1.0, 0.0],
                                            [0.0, 0.0, 0.0]]), pixel_scale=1.0)
            im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
            li = lensing_image.LensingImage(im, ma, sub_grid_size=2)

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)))
            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li.grids], border=None)
            fit = lensing_fitting.LensingInversionFit(lensing_images=[li], tracer=tracer)

            cov_matrix = np.array([[1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]])
            reg_matrix = np.array([[2.0, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                   [-1.0, 3.0, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                   [0.0, -1.0, 2.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                   [-1.0, 0.0, 0.0, 3.0, -1.0, 0.0, -1.0, 0.0, 0.0],
                                   [0.0, -1.0, 0.0, -1.0, 4.0, -1.0, 0.0, -1.0, 0.0],
                                   [0.0, 0.0, -1.0, 0.0, -1.0, 3.0, 0.0, 0.0, - 1.0],
                                   [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.0, -1.0, 0.0],
                                   [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.0, -1.0],
                                   [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.0]])
            reg_matrix = reg_matrix + 1e-8 * np.identity(9)
            cov_reg_matrix = cov_matrix + reg_matrix

            chi_sq_term = 0.0
            gl_term = 1e-8
            det_cov_reg_term = np.log(np.linalg.det(cov_reg_matrix))
            det_reg_term = np.log(np.linalg.det(reg_matrix))
            noise_term = 9.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            evidence = -0.5 * (chi_sq_term + gl_term + det_cov_reg_term - det_reg_term + noise_term)

            assert fit.evidence == pytest.approx(evidence, 1e-4)

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g0],
                                                         image_plane_grids=[li_manual.grids],
                                                         border=None)

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_manual, tracer=tracer)

            mapper = pix.mapper_from_grids_and_border(li_manual.grids, None)
            inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper, regularization=reg,
                                                                                          image=li_manual, noise_map=li_manual.noise_map_,
                                                                                          convolver=li_manual.convolver_mapping_matrix)

            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_manual],
                                                                     model_datas_=[inversion.reconstructed_data_vector])
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_manual.noise_map_])

            assert li_manual.grids.regular.scaled_array_from_array_1d(li_manual.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(inversion.reconstructed_data_vector) == \
                   pytest.approx(fit.model_data, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_manual.noise_map_])
            likelihoods_with_regularization = \
                fitting_util.likelihoods_with_regularization_from_chi_squared_regularization_and_noise_terms(
                    chi_squared_terms, [inversion.regularization_term], noise_terms)

            assert likelihoods_with_regularization[0] == pytest.approx(fit.likelihoods_with_regularization[0], 1e-4)

            evidences = fitting_util.evidences_from_reconstruction_terms(chi_squared_terms, [inversion.regularization_term],
                                                                    [inversion.log_det_curvature_reg_matrix_term],
                                                                    [inversion.log_det_regularization_matrix_term],
                                                                    noise_terms)

            assert evidences[0] == fit.evidences[0]

            fast_evidence = lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(lensing_image=li_manual,
                                                                                          tracer=tracer)
            assert fast_evidence == fit.evidences[0]


class TestHyperLensingInversionFit:

    class TestRectangularInversion:

        def test__image_all_1s__direct_image_to_source_mapping__perfect_fit_even_with_regularization(self):

            im = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 0.0, 0.0, 0.0, 0.0]]).view(image.Image)
            ma = np.array([[True, True, True, True, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, True, True, True, True]])
            ma = mask.Mask(ma, pixel_scale=1.0)
            psf = image.PSF(array=np.array([[0.0, 0.0, 0.0],
                                            [0.0, 1.0, 0.0],
                                            [0.0, 0.0, 0.0]]), pixel_scale=1.0)
            im = image.Image(im, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
            hyper_model_image = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
            hyper_galaxy_images = [np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])]
            li = lensing_image.LensingHyperImage(im, ma, hyper_model_image=hyper_model_image,
                                                 hyper_galaxy_images=hyper_galaxy_images,
                                                 hyper_minimum_values=[0.0, 0.0],
                                                 sub_grid_size=2)

            hyper_galaxy = g.HyperGalaxy(contribution_factor=0.0, noise_factor=1.0, noise_power=1.0)

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)),
                                  hyper_galaxy=hyper_galaxy)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li.grids], border=None)

            fit = lensing_fitting.HyperLensingInversionFit(lensing_hyper_images=[li], tracer=tracer)

            curvature_matrix = np.array([[0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                         [0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                         [0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                         [0.0, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0],
                                         [0.0, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0],
                                         [0.0, 0.0, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0],
                                         [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0],
                                         [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25, 0.0],
                                         [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25]])
            regularization_matrix = np.array([[2.0, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                              [-1.0, 3.0, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                                              [0.0, -1.0, 2.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                                              [-1.0, 0.0, 0.0, 3.0, -1.0, 0.0, -1.0, 0.0, 0.0],
                                              [0.0, -1.0, 0.0, -1.0, 4.0, -1.0, 0.0, -1.0, 0.0],
                                              [0.0, 0.0, -1.0, 0.0, -1.0, 3.0, 0.0, 0.0, - 1.0],
                                              [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.0, -1.0, 0.0],
                                              [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.0, -1.0],
                                              [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.0]])
            regularization_matrix = regularization_matrix + 1e-8 * np.identity(9)
            curvature_reg_matrix = curvature_matrix + regularization_matrix

            scaled_chi_sq_term = 0.0
            gl_term = 1e-8
            det_curvature_reg_term = np.log(np.linalg.det(curvature_reg_matrix))
            det_regularization_term = np.log(np.linalg.det(regularization_matrix))
            scaled_noise_term = 9.0 * np.log(2 * np.pi * 2.0 ** 2.0)

            scaled_evidence = -0.5 * (scaled_chi_sq_term + gl_term + det_curvature_reg_term - det_regularization_term +
                                      scaled_noise_term)

            assert fit.scaled_evidences == pytest.approx(scaled_evidence, 1e-4)

    class TestAbstractLogic:

        def test__logic_in_abstract_fit(self, li_hyper_no_blur):
            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li_hyper_no_blur.grids],
                                                         border=None)

            fit = lensing_fitting.HyperLensingInversionFit(lensing_hyper_images=[li_hyper_no_blur], tracer=tracer)

            assert fit.total_inversions == 1

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_hyper_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            mapper = pix.mapper_from_grids_and_border(li_hyper_manual.grids, li_hyper_manual.border)
            reg = regularization.Constant(coefficients=(1.0,))

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
            hyp_galaxy = g.Galaxy(hyper_galaxy=hyper_galaxy)
            inv_galaxy = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[hyp_galaxy, hyp_galaxy],
                                                         source_galaxies=[inv_galaxy],
                                                         image_plane_grids=[li_hyper_manual.grids],
                                                         border=None)

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_hyper_manual, tracer=tracer)

            inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper,
                                                                                          regularization=reg, image=li_hyper_manual, noise_map=li_hyper_manual.noise_map_,
                                                                                          convolver=li_hyper_manual.convolver_mapping_matrix)

            residuals = fitting_util.residuals_from_datas_and_model_datas([li_hyper_manual],
                                                                     [inversion.reconstructed_data_vector])
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [li_hyper_manual.noise_map_])

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(li_hyper_manual.noise_map_) == \
                   pytest.approx(fit.noise_maps[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(inversion.reconstructed_data_vector) == \
                   pytest.approx(fit.model_datas[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residuals[0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squareds[0], 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_hyper_manual.noise_map_])

            likelihoods_with_regularization = \
                fitting_util.likelihoods_with_regularization_from_chi_squared_regularization_and_noise_terms(chi_squared_terms,
                                                                                                        [inversion.regularization_term], noise_terms)

            assert likelihoods_with_regularization[0] == pytest.approx(fit.likelihoods_with_regularization[0], 1e-2)

            evidences = fitting_util.evidences_from_reconstruction_terms(chi_squared_terms, [inversion.regularization_term],
                                                                    [inversion.log_det_curvature_reg_matrix_term],
                                                                    [inversion.log_det_regularization_matrix_term],
                                                                    noise_terms)

            assert evidences[0] == fit.evidences[0]

            contributions = fitting_util.contributions_from_hyper_images_and_galaxies(li_hyper_manual.hyper_model_image,
                                                                                 li_hyper_manual.hyper_galaxy_images, [hyper_galaxy, hyper_galaxy],
                                                                                 li_hyper_manual.hyper_minimum_values)

            scaled_noise_maps = fitting_util.scaled_noises_from_fitting_hyper_images_contributions_and_hyper_galaxies(
                fitting_hyper_images=[li_hyper_manual], contributions_=[contributions], hyper_galaxies=[hyper_galaxy,
                                                                                                        hyper_galaxy])

            scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper,
                                                                                                 regularization=reg, image=li_hyper_manual, noise_map=scaled_noise_maps[0],
                                                                                                 convolver=li_hyper_manual.convolver_mapping_matrix)

            scaled_residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_hyper_manual],
                                                                            model_datas_=[scaled_inversion.reconstructed_data_vector])

            scaled_chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(scaled_residuals,
                                                                                     scaled_noise_maps)

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[0]) == \
                   pytest.approx(fit.contributions[0][0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[1]) == \
                   pytest.approx(fit.contributions[0][1], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_noise_maps[0]) == \
                   pytest.approx(fit.scaled_noise_map, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_residuals[0]) == \
                   pytest.approx(fit.scaled_residual, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_chi_squareds[0]) == \
                   pytest.approx(fit.scaled_chi_squared, 1e-4)

            scaled_chi_squared_terms= fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=scaled_chi_squareds)
            scaled_noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=scaled_noise_maps)

            scaled_evidences = fitting_util.evidences_from_reconstruction_terms(scaled_chi_squared_terms,
                                                                           [scaled_inversion.regularization_term], [scaled_inversion.log_det_curvature_reg_matrix_term],
                                                                           [scaled_inversion.log_det_regularization_matrix_term], scaled_noise_terms)
            assert scaled_evidences[0] == fit.scaled_evidences[0]

            fast_scaled_evidence = \
                lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(lensing_image=li_hyper_manual,
                                                                              tracer=tracer)
            assert fast_scaled_evidence == scaled_evidences[0]


class TestLensingProfileInversionFit:

    class TestModelImagesOfPLanes:

        def test___model_images_of_planes_are_profile_and_inversion_images(self, li_manual):

            galaxy_light = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li_manual.grids],
                                                         border=None)

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_manual, tracer=tracer)

            assert (fit.model_images_of_planes[0][0] == fit.profile_model_images[0]).all()
            assert (fit.model_images_of_planes[0][1] == fit.inversion_model_images[0]).all()

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_manual):

            galaxy_light = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light], source_galaxies=[galaxy_pix],
                                                         image_plane_grids=[li_manual.grids],
                                                         border=None)

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_manual, tracer=tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            profile_model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im,
                                                                                  blurring_images_=blurring_im, convolvers=[li_manual.convolver_image])
            profile_subtracted_images = [li_manual[:] - profile_model_images[0]]

            assert li_manual.grids.regular.scaled_array_from_array_1d(profile_model_images[0]) == \
                   pytest.approx(fit.profile_model_image, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(profile_subtracted_images[0]) == \
                   pytest.approx(fit.profile_subtracted_image, 1e-4)

            mapper = pix.mapper_from_grids_and_border(li_manual.grids, None)
            inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper, regularization=reg,
                                                                                          image=profile_subtracted_images[0],
                                                                                          noise_map=li_manual.noise_map_,
                                                                                          convolver=li_manual.convolver_mapping_matrix)

            model_images = [profile_model_images[0] + inversion.reconstructed_data_vector]
            residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_manual], model_datas_=model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals,
                                                                              noise_maps_=[li_manual.noise_map_])

            assert li_manual.grids.regular.scaled_array_from_array_1d(li_manual.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(inversion.reconstructed_data_vector) == \
                   pytest.approx(fit.inversion_model_image, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) == \
                   pytest.approx(fit.model_image, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) ==\
                   pytest.approx(fit.residual, 1e-4)
            assert li_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps(noise_maps_=[li_manual.noise_map_])
            evidences = fitting_util.evidences_from_reconstruction_terms(chi_squared_terms,
                                                                    [inversion.regularization_term],
                                                                    [inversion.log_det_curvature_reg_matrix_term],
                                                                    [inversion.log_det_regularization_matrix_term],
                                                                    noise_terms)

            assert evidences[0] == fit.evidences[0]

            fast_evidence = lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(lensing_image=li_manual,
                                                                                          tracer=tracer)
            assert fast_evidence == evidences[0]


class TestHyperLensingProfileInversionFit:

    class TestCompareToManual:

        def test___manual_image_and_psf(self, li_hyper_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
            hyp_galaxy = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)
            inv_galaxy = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[hyp_galaxy, hyp_galaxy],
                                                         source_galaxies=[inv_galaxy],
                                                         image_plane_grids=[li_hyper_manual.grids],
                                                         border=None)

            fit = lensing_fitting.fit_lensing_image_with_tracer(lensing_image=li_hyper_manual, tracer=tracer)

            image_im = tracer.image_plane_images_
            blurring_im = tracer.image_plane_blurring_images_
            profile_model_images = fitting_util.blur_images_including_blurring_regions(images_=image_im,
                                                                                  blurring_images_=blurring_im, convolvers=[li_hyper_manual.convolver_image])
            profile_subtracted_images = [li_hyper_manual[:] - profile_model_images[0]]

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(profile_model_images[0]) == \
                   pytest.approx(fit.profile_model_image, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(profile_subtracted_images[0]) == \
                   pytest.approx(fit.profile_subtracted_image, 1e-4)

            mapper = pix.mapper_from_grids_and_border(li_hyper_manual.grids, li_hyper_manual.border)
            inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper, regularization=reg,
                                                                                          image=profile_subtracted_images[0], noise_map=li_hyper_manual.noise_map_,
                                                                                          convolver=li_hyper_manual.convolver_mapping_matrix)

            model_images = [profile_model_images[0] + inversion.reconstructed_data_vector]
            residuals = fitting_util.residuals_from_datas_and_model_datas([li_hyper_manual], model_images)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [li_hyper_manual.noise_map_])

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(li_hyper_manual.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(inversion.reconstructed_data_vector) == \
                   pytest.approx(fit.inversion_model_image, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(model_images[0]) ==\
                   pytest.approx(fit.model_data, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) ==\
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps([li_hyper_manual.noise_map_])
            evidences = fitting_util.evidences_from_reconstruction_terms(chi_squared_terms, [inversion.regularization_term],
                                                                    [inversion.log_det_curvature_reg_matrix_term],
                                                                    [inversion.log_det_regularization_matrix_term],
                                                                    noise_terms)

            assert evidences[0] == fit.evidences[0]

            contributions = fitting_util.contributions_from_hyper_images_and_galaxies(li_hyper_manual.hyper_model_image,
                                                                                 li_hyper_manual.hyper_galaxy_images, [hyper_galaxy, hyper_galaxy],
                                                                                 li_hyper_manual.hyper_minimum_values)

            scaled_noise_maps = fitting_util.scaled_noises_from_fitting_hyper_images_contributions_and_hyper_galaxies(
                fitting_hyper_images=[li_hyper_manual], contributions_=[contributions], hyper_galaxies=[hyper_galaxy,
                                                                                                        hyper_galaxy])

            scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(mapper=mapper,
                                                                                                 regularization=reg, image=profile_subtracted_images[0], noise_map=scaled_noise_maps[0],
                                                                                                 convolver=li_hyper_manual.convolver_mapping_matrix)

            scaled_model_images = [profile_model_images[0] + scaled_inversion.reconstructed_data_vector]

            scaled_residuals = fitting_util.residuals_from_datas_and_model_datas(datas_=[li_hyper_manual],
                                                                            model_datas_=scaled_model_images)

            scaled_chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(scaled_residuals,
                                                                                     scaled_noise_maps)

            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[0]) == \
                   pytest.approx(fit.contributions[0][0], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(contributions[1]) == \
                   pytest.approx(fit.contributions[0][1], 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_noise_maps[0]) == \
                   pytest.approx(fit.scaled_noise_map, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_model_images[0]) == \
                   pytest.approx(fit.scaled_model_image, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_residuals[0]) == \
                   pytest.approx(fit.scaled_residual, 1e-4)
            assert li_hyper_manual.grids.regular.scaled_array_from_array_1d(scaled_chi_squareds[0]) == \
                   pytest.approx(fit.scaled_chi_squared, 1e-4)

            scaled_chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(scaled_chi_squareds)
            scaled_noise_terms = fitting_util.noise_terms_from_noise_maps(scaled_noise_maps)
            scaled_evidence = fitting_util.evidences_from_reconstruction_terms(
                scaled_chi_squared_terms, [scaled_inversion.regularization_term],
                [scaled_inversion.log_det_curvature_reg_matrix_term],
                [scaled_inversion.log_det_regularization_matrix_term],
                scaled_noise_terms)

            assert scaled_evidence[0] == fit.scaled_evidences[0]

            fast_scaled_evidence = lensing_fitting.fast_likelihood_from_lensing_image_and_tracer(
                lensing_image=li_hyper_manual, tracer=tracer)

            assert fast_scaled_evidence == scaled_evidence[0]


class MockTracerPositions:

    def __init__(self, positions, noise=None):
        self.positions = positions
        self.noise = noise


class TestPositionFit:

    def test__x1_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(2)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 3.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(1.0) + np.square(3.0))

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [-9.0, -4.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0], [0.1, 0.1], [-0.1, -0.1],
                                                          [0.3, 0.4], [-0.6, 0.5]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [8.0, 4.0], [-9.0, -4.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_sets_of_positions__multiple_sets_of_max_distances(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                                                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                                                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)

        assert fit.maximum_separations[0] == 1.0
        assert fit.maximum_separations[1] == np.sqrt(18)
        assert fit.maximum_separations[2] == np.sqrt(18)

    def test__likelihood__is_sum_of_separations_divided_by_noise(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                                                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                                                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)
        assert fit.chi_squareds[0] == 1.0
        assert fit.chi_squareds[1] == pytest.approx(18.0, 1e-4)
        assert fit.chi_squareds[2] == pytest.approx(18.0, 1e-4)
        assert fit.likelihood == pytest.approx(-0.5 * (1.0 + 18 + 18), 1e-4)

        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=2.0)
        assert fit.chi_squareds[0] == (1.0 / 2.0) ** 2.0
        assert fit.chi_squareds[1] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0, 1e-4)
        assert fit.chi_squareds[2] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0, 1e-4)
        assert fit.likelihood == pytest.approx(-0.5 * ((1.0 / 2.0) ** 2.0 + (np.sqrt(18.0) / 2.0) ** 2.0 +
                                                       (np.sqrt(18.0) / 2.0) ** 2.0), 1e-4)

    def test__threshold__if_not_met_returns_ray_tracing_exception(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lensing_fitting.PositionFit(positions=tracer.positions, noise=1.0)

        assert fit.maximum_separation_within_threshold(threshold=100.0) == True
        assert fit.maximum_separation_within_threshold(threshold=0.1) == False



from astropy.io import fits
import numpy as np
import os

path = '{}/../test_files/array/'.format(os.path.dirname(os.path.realpath(__file__)))

array1 = np.ones((3,3))
array2 = 2.0*np.ones((3,3))
array3 = 3.0*np.ones((3,3))
array4 = 4.0*np.ones((3,3))
array5 = 5.0*np.ones((3,3))
array6 = 6.0*np.ones((3,3))
array7 = 7.0*np.ones((3,3))
array8 = 8.0*np.ones((3,3))

fits.writeto(data=array2, filename=path + '3x3_twos.fits')
fits.writeto(data=array3, filename=path + '3x3_threes.fits')
fits.writeto(data=array4, filename=path + '3x3_fours.fits')
fits.writeto(data=array5, filename=path + '3x3_fives.fits')
fits.writeto(data=array6, filename=path + '3x3_sixes.fits')
fits.writeto(data=array7, filename=path + '3x3_sevens.fits')
fits.writeto(data=array8, filename=path + '3x3_eights.fits')
stop
new_hdul = fits.HDUList()
new_hdul.append(fits.ImageHDU(array1))
new_hdul.append(fits.ImageHDU(array2))
new_hdul.append(fits.ImageHDU(array3))
new_hdul.append(fits.ImageHDU(array4))
new_hdul.append(fits.ImageHDU(array5))
new_hdul.append(fits.ImageHDU(array6))
new_hdul.append(fits.ImageHDU(array7))
new_hdul.append(fits.ImageHDU(array8))

new_hdul.writeto(path+'3x3_multiple_hdu.fits')


import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids, mask
from autolens.model.inversion import inversions
from test.mock.mock_inversion import MockConvolver


class MockGeometry(object):

    def __init__(self):

        self.pixel_scales = (1.0, 1.0)
        self.origin = (0.0, 0.0)
        self.pixel_centres = None
        self.pixel_neighbors = [[]]
        self.pixel_neighbors_size = []


class MockMapper(object):

    def __init__(self, matrix_shape, grids=None):

        self.grids = grids
        self.mapping_matrix = np.ones(matrix_shape)
        self.geometry = MockGeometry()


class MockRegularization(object):

    def __init__(self, matrix_shape):
        self.shape = matrix_shape

    def regularization_matrix_from_pixel_neighbors(self, pixel_neighbors, pixel_neighbors_size):
        return np.array([[1.0, 0.0, 0.0],
                         [0.0, 1.0, 0.0],
                         [0.0, 0.0, 1.0]])


class TestRegularizationTerm:

    def test__solution_all_1s__regularization_matrix_simple(self):

        matrix_shape = (3,3)

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape), regularization=MockRegularization(matrix_shape))

        inv.solution_vector = np.array([1.0, 1.0, 1.0])

        inv.regularization_matrix = np.array([[1.0, 0.0, 0.0],
                                              [0.0, 1.0, 0.0],
                                              [0.0, 0.0, 1.0]])

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [1.0, 1.0, 1.0] * [1.0, 1.0, 1.0] = [(1.0*1.0) + (1.0*0.0) + (1.0*0.0)] = [1.0, 1.0, 1.0]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*1.0) + (1.0*0.0)]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*0.0) + (1.0*1.0)]

        # (s_T * H) * s = [1.0, 1.0, 1.0] * [1.0] = 3.0
        #                                   [1.0]
        #                                   [1.0]

        assert inv.regularization_term == 3.0

    def test__solution_and_regularization_matrix_range_of_values(self):

        matrix_shape = (3,3)

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape), regularization=MockRegularization(matrix_shape))

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [2.0, 3.0, 5.0] * [2.0,  -1.0,  0.0] = [(2.0* 2.0) + (3.0*-1.0) + (5.0 *0.0)] = [1.0, -1.0, 7.0]
        #                             [-1.0,  2.0, -1.0]   [(2.0*-1.0) + (3.0* 2.0) + (5.0*-1.0)]
        #                             [ 0.0, -1.0,  2.0]   [(2.0* 0.0) + (3.0*-1.0) + (5.0 *2.0)]

        # (s_T * H) * s = [1.0, -1.0, 7.0] * [2.0] = 34.0
        #                                    [3.0]
        #                                    [5.0]

        inv.solution_vector = np.array([2.0, 3.0, 5.0])

        inv.regularization_matrix = np.array([[2.0, -1.0, 0.0],
                                          [-1.0, 2.0, -1.0],
                                          [0.0, -1.0, 2.0]])

        assert inv.regularization_term == 34.0


class TestLogDetMatrix:

    def test__determinant_of_positive_definite_matrix_via_cholesky(self):

        matrix_shape = (3,3)

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape), regularization=MockRegularization(matrix_shape))

        matrix = np.array([[1.0, 0.0, 0.0],
                           [0.0, 1.0, 0.0],
                           [0.0, 0.0, 1.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        assert log_determinant == pytest.approx(inv.log_determinant_of_matrix_cholesky(matrix), 1e-4)

    def test__determinant_of_positive_definite_matrix_2_via_cholesky(self):

        matrix_shape = (3,3)

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape), regularization=MockRegularization(matrix_shape))

        matrix = np.array([[2.0, -1.0, 0.0],
                           [-1.0, 2.0, -1.0],
                           [0.0, -1.0, 2.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        assert log_determinant == pytest.approx(inv.log_determinant_of_matrix_cholesky(matrix), 1e-4)

    def test__matrix_not_positive_definite__raises_reconstruction_exception(self):

        matrix_shape = (3,3)

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape), regularization=MockRegularization(matrix_shape))

        matrix = np.array([[2.0, 0.0, 0.0],
                           [-1.0, 2.0, -1.0],
                           [0.0, -1.0, 0.0]])

        with pytest.raises(exc.InversionException):
            assert pytest.approx(inv.log_determinant_of_matrix_cholesky(matrix), 1e-4)


class TestReconstructedDataVectorAndImage:

    def test__solution_all_1s__simple_blurred_mapping_matrix__correct_reconstructed_image(self):

        matrix_shape = (3,3)

        msk = mask.Mask(array=np.array([[True, True, True],
                                        [False, False, False],
                                        [True, True, True]]), pixel_scale=1.0)

        imaging_grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=msk, sub_grid_size=1,
                                                                                    psf_shape=(1,1))

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape, imaging_grids),
                                   regularization=MockRegularization(matrix_shape))

        inv.solution_vector = np.array([1.0, 1.0, 1.0, 1.0])

        inv.blurred_mapping_matrix = np.array([[1.0, 1.0, 1.0, 1.0],
                                           [1.0, 0.0, 1.0, 1.0],
                                           [1.0, 0.0, 0.0, 0.0]])
        # Image pixel 0 maps to 4 pixs pixxels -> value is 4.0
        # Image pixel 1 maps to 3 pixs pixxels -> value is 3.0
        # Image pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (inv.reconstructed_data_vector == np.array([4.0, 3.0, 1.0])).all()
        assert (inv.reconstructed_data == np.array([[0.0, 0.0, 0.0],
                                                    [4.0, 3.0, 1.0],
                                                    [0.0, 0.0, 0.0]]))

    def test__solution_different_values__simple_blurred_mapping_matrix__correct_reconstructed_image(self):

        matrix_shape = (3,3)

        msk = mask.Mask(array=np.array([[True, True, True],
                                        [False, False, False],
                                        [True, True, True]]), pixel_scale=1.0)

        imaging_grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=msk, sub_grid_size=1,
                                                                                    psf_shape=(1,1))

        inv = inversions.Inversion(image=np.ones(9), noise_map=np.ones(9), convolver=MockConvolver(matrix_shape),
                                   mapper=MockMapper(matrix_shape, imaging_grids), regularization=MockRegularization(matrix_shape))

        inv.solution_vector = np.array([1.0, 2.0, 3.0, 4.0])

        inv.blurred_mapping_matrix = np.array([[1.0, 1.0, 1.0, 1.0],
                                               [1.0, 0.0, 1.0, 1.0],
                                               [1.0, 0.0, 0.0, 0.0]])

        # # Image pixel 0 maps to 4 pixs pixxels -> value is 1.0 + 2.0 + 3.0 + 4.0 = 10.0
        # # Image pixel 1 maps to 3 pixs pixxels -> value is 1.0 + 3.0 + 4.0
        # # Image pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (inv.reconstructed_data_vector == np.array([10.0, 8.0, 1.0])).all()
        assert (inv.reconstructed_data == np.array([[0.0, 0.0, 0.0],
                                                    [10.0, 8.0, 1.0],
                                                    [0.0, 0.0, 0.0]]))


from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.model.galaxy import galaxy
from autolens.lensing import ray_tracing
from autolens.data.array import grids
from autolens.lensing.plotters import plane_plotters
from autolens.lensing.plotters import ray_tracing_plotters

# In the previous example, we used light-profiles, mass-profiles, galaxies and planes to create an regular-plane +
# source-plane strong lens system. However, this took a lot of lines of code to do, so in this example we'll use the
# 'ray-tracing module to do it a lot faster!

# Lets use the same grid as always, you should be used to seeing this now!
image_plane_grids = grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05,
                                                               sub_grid_size=2)

# Unlike the previous tutorial, we'll offset our lens model_galaxy slightly from the source, to get a slightly ray-tracing
# paths.
sis_mass_profile = mass_profiles.SphericalIsothermal(centre=(0.1, 0.1), einstein_radius=1.6)
lens_galaxy = galaxy.Galaxy(mass=sis_mass_profile)
print(lens_galaxy)

# We'll also make our source model_galaxy ellipctical and increase its Sersic index.
sersic_light_profile = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0,
                                                       intensity=1.0, effective_radius=1.0, sersic_index=2.5)
source_galaxy = galaxy.Galaxy(light=sersic_light_profile)
print(source_galaxy)

# Finally, we can use a 'tracer_without_subhalo', the lens model_galaxy and the source model_galaxy to ray-trace our grids. When we pass our
# galaxies and grids into the Tracer below, the following happens:

# 1) Using the lens-model_galaxy's mass-profile, the deflection angle of every regular-plane grid coordinate is computed.
# 2) These deflection angles are used to trace every regular-plane coordinate to a source-plane coordinate.
# 3) This creates a source-plane grid of lensed coordinates.

# This is what we did using Plane's in the preious chapter, but its a lot less lines of code!
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                             image_plane_grids=[image_plane_grids])

# The tracer is composed of an image-plane and source-plane, just like in the previous example!
print('Image-pixel 1 regular-plane coordinate')
print(tracer.image_plane.grids[0].regular[0])
print('Image-pixel 2 regular-plane coordinate')
print(tracer.image_plane.grids[0].regular[1])
print('Image-pixel 3 regular-plane coordinate')
print(tracer.image_plane.grids[0].regular[2])

# And the source-plane's grid has been deflected.
print('Image-pixel 1 source-plane coordinate')
print(tracer.source_plane.grids[0].regular[0])
print('Image-pixel 2 source-plane coordinate')
print(tracer.source_plane.grids[0].regular[1])
print('Image-pixel 3 source-plane coordinate')
print(tracer.source_plane.grids[0].regular[2])

# We can use the plane_plotter to plot these grids, like before.
plane_plotters.plot_plane_grid(plane=tracer.image_plane, title='Image-plane Grid')
plane_plotters.plot_plane_grid(plane=tracer.source_plane, title='Source-plane Grid')
plane_plotters.plot_plane_grid(plane=tracer.source_plane, axis_limits=[-0.1, 0.3, -0.1, 0.3], title='Source-plane Grid')

# AutoLens has tools for plotting a tracer_without_subhalo. This plots the following:
# 1) The regular-plane regular, computed by tracing the source model_galaxy's light 'forwards' through the tracer_without_subhalo.
# 2) The source-plane regular, showing the source model_galaxy's true appearance (i.e. if it were not lensed).
# 3) The regular-plane surface density, computed using the lens model_galaxy's mass profile.
# 4) The regular-plane gravitational potential, computed using the lens model_galaxy's mass profile.
# 5) The regular-plane deflection angles, computed using the lens model_galaxy's mass profile.
ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer)

# These attributes can be assessed by print statements (Which you might notice have been converted to 2D NumPy arrays
# which are the same dimensions as our input regular!).
print('Image-Pixel 1 - Tracer - Surface Density:')
print(tracer.surface_density[0,0])
print('Image-Pixel 2 - Tracer - Surface Density:')
print(tracer.surface_density[0,1])
print('Image-Pixel 3 - Tracer - Surface Density:')
print(tracer.surface_density[0,2])
print('Image-Pixel 101 - Tracer - Surface Density:')
print(tracer.surface_density[1,0])

# Of course, these surface densities are identical to the regular-plane surface densities, as it's only the lens model_galaxy
# that contributes to the overall mass of the system.
print('Image-Pixel 1 - Image-Plane - Surface Density:')
print(tracer.image_plane.surface_density[0,0])
print('Image-Pixel 2 - Image-Plane - Surface Density:')
print(tracer.image_plane.surface_density[0,1])
print('Image-Pixel 3 - Image-Plane - Surface Density:')
print(tracer.image_plane.surface_density[0,2])
print('Image-Pixel 101 - Image-Plane - Surface Density:')
print(tracer.image_plane.surface_density[1,0])

# I've left the rest below commented to avoid too much printing, but you can inspect their datas if you please!
# print('Potential:')
# print(tracer_without_subhalo.potential)
# print(tracer_without_subhalo.image_plane.potential)
# print('Deflections:')
# print(tracer_without_subhalo.deflections_x)
# print(tracer_without_subhalo.deflections_y)
# print(tracer_without_subhalo.image_plane.deflections_x)
# print(tracer_without_subhalo.image_plane.deflections_y)
# print('Image-plane Image:')
# print(tracer_without_subhalo.image_plane_image)
# print(tracer_without_subhalo.image_plane.image_plane_image)
# print('Source-plane Image:')
# print(tracer_without_subhalo.source_plane_image)
# print(tracer_without_subhalo.image_plane.source_plane_image)

# You can also plot the above attributes on individual figures, using appropriate ray-tracing plotter (I've left most
# commented out again for convinience)
ray_tracing_plotters.plot_surface_density(tracer=tracer)
# ray_tracing_plotters.plot_potential(tracer_normal=tracer_normal)
# ray_tracing_plotters.plot_deflections_y(tracer_normal=tracer_normal)
#ray_tracing_plotters.plot_deflections_x(tracer_normal=tracer_normal)
# ray_tracing_plotters.plot_image_plane_image(tracer_normal=tracer_normal)

# Before we finish, the attentive amongst you might be wondering 'why do both the regular-plane and tracer_without_subhalo have the
# attributes surface density / potential / deflection angles, when, by definition, the two are identical'. Think about
# it - only mass profiles contribute to these quantities, and only the regular-plane should have galaxies with
# mass profiles! There are two reasons:

# 1) Convinience - You could always write 'tracer_without_subhalo.image_plane.surface_density' and
#                  'plane_plotters.surface_density(plane=tracer_without_subhalo.image_plane). However, code appears neater if you can
#                   just write 'tracer_normal.surface_density' and 'ray_tracing_plotters.plot_surface_density(tracer_normal=tracer_normal).

# 2) Multi-plane lensing - For now, we're focused on the regular-plane + source-plane lensing configuration. However,
#                          there are strong lens system where there are more than 2 planes! In these instances, the
#                          surface density, potential and deflections of each plane is different to that of the tracer_without_subhalo.
#                          This is way beyond the scope of this chapter, but be reassured that what you're learning now
#                          will prepare you for the advanced chapters later on!

# And with that, we're done. You've performed your first actual lensing with PyAutoLens!

# 1) Change the lens model_galaxy's einstein radius - what happens to the tracer_without_subhalo's regular-plane regular?
# 2) Change the source model_galaxy's effective radius - how does the regular-plane regular's appearance change?
# 3) Experiment with different light-profiles and mass-profiles. In particular, change the SphericalIsothermal
#    mass-profile to an EllipticalIsothermal mass-profile and set its axis_ratio parameter to 0.8. What
#    happens to the number of source regular?



import os

import pytest
from autofit import conf
from autofit.core import model_mapper as mm
from autofit.core import non_linear

from autolens.model.galaxy import galaxy as g, galaxy_model as gp
from autolens.pipeline import phase as ph
from autolens.pipeline import phase_property

directory = os.path.dirname(os.path.realpath(__file__))

conf.instance = conf.Config("{}/../../workspace/config".format(directory),
                            "{}/../../workspace/output/".format(directory))


class NLO(non_linear.NonLinearOptimizer):
    def fit(self, analysis):
        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(instance)
                self.result = non_linear.Result(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)
        fitness_function(self.variable.prior_count * [0.5])

        return fitness_function.result


@pytest.fixture(name='phase')
def make_phase():
    class MyPhase(ph.Phase):
        prop = phase_property.PhaseProperty("prop")

    return MyPhase(optimizer_class=NLO)


@pytest.fixture(name='list_phase')
def make_list_phase():
    class MyPhase(ph.Phase):
        prop = phase_property.PhasePropertyCollection("prop")

    return MyPhase(optimizer_class=NLO)


class TestPhaseProperty(object):
    def test_phase_property(self, phase):
        phase.prop = gp.GalaxyModel()

        assert phase.variable.prop == phase.prop

        galaxy = g.Galaxy()
        phase.prop = galaxy

        assert phase.constant.prop == galaxy
        assert not hasattr(phase.variable, "prop")

        phase.prop = gp.GalaxyModel()
        assert not hasattr(phase.constant, "prop")


class TestPhasePropertyList(object):
    def test_constants(self, list_phase):
        objects = [g.Galaxy(), g.Galaxy()]

        list_phase.prop = objects

        assert list_phase.constant.prop == objects
        assert len(list_phase.variable.prop) == 0

        assert list_phase.prop == objects

    def test_classes(self, list_phase):
        objects = [gp.GalaxyModel(), gp.GalaxyModel()]

        list_phase.prop = objects

        assert list_phase.variable.prop == objects
        assert len(list_phase.constant.prop) == 0

        assert list_phase.prop == objects

    def test_abstract_prior_models(self, list_phase):
        objects = [mm.AbstractPriorModel(), mm.AbstractPriorModel()]

        list_phase.prop = objects

        assert list_phase.variable.prop == objects
        assert len(list_phase.constant.prop) == 0

        assert list_phase.prop == objects

    def test_mix(self, list_phase):
        objects = [gp.GalaxyModel(), g.Galaxy()]

        list_phase.prop = objects

        assert list_phase.variable.prop == [objects[0]]
        assert list_phase.constant.prop == [objects[1]]

        assert list_phase.prop == objects

    def test_set_item(self, list_phase):
        galaxy_prior_0 = gp.GalaxyModel()
        objects = [galaxy_prior_0, g.Galaxy()]

        list_phase.prop = objects
        assert_ordered(list_phase.prop)

        galaxy_prior_1 = gp.GalaxyModel()
        list_phase.prop[1] = galaxy_prior_1

        assert_ordered(list_phase.prop)

        assert list_phase.constant.prop == []
        assert list_phase.variable.prop == [galaxy_prior_0, galaxy_prior_1]

        galaxy = g.Galaxy()

        list_phase.prop[0] = galaxy

        assert_ordered(list_phase.prop)

        assert list_phase.prop == [galaxy, galaxy_prior_1]

        assert list_phase.constant.prop == [galaxy]
        assert list_phase.variable.prop == [galaxy_prior_1]


class TestPhasePropertyCollectionAttributes(object):
    def test_set_list_as_dict(self, list_phase):
        galaxy_model = gp.GalaxyModel()
        list_phase.prop = dict(one=galaxy_model)

        assert len(list_phase.prop) == 1
        # noinspection PyUnresolvedReferences
        assert list_phase.prop.one == galaxy_model

    def test_override_property(self, list_phase):
        galaxy_model = gp.GalaxyModel()

        list_phase.prop = dict(one=gp.GalaxyModel())

        list_phase.prop.one = galaxy_model

        assert len(list_phase.prop) == 1
        assert list_phase.prop.one == galaxy_model

    def test_named_list_items(self, list_phase):
        galaxy_model = gp.GalaxyModel()
        list_phase.prop = [galaxy_model]

        # noinspection PyUnresolvedReferences
        assert list_phase.prop.prop_0 == galaxy_model

    def test_mix(self, list_phase):
        objects = dict(one=gp.GalaxyModel(), two=g.Galaxy())

        list_phase.prop = objects

        assert list_phase.variable.prop == [objects["one"]]
        assert list_phase.constant.prop == [objects["two"]]

        list_phase.prop.one = g.Galaxy()

        assert len(list_phase.variable.prop) == 0
        assert len(list_phase.constant.prop) == 2

    def test_named_attributes_in_variable(self, list_phase):
        galaxy_model = gp.GalaxyModel(variable_redshift=True)
        list_phase.prop = dict(one=galaxy_model)

        assert list_phase.variable.prior_count == 1
        assert list_phase.variable.one == galaxy_model

        instance = list_phase.variable.instance_from_prior_medians()

        assert instance.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_variable_override(self, list_phase):
        galaxy_model = gp.GalaxyModel(variable_redshift=True)
        list_phase.prop = dict(one=gp.GalaxyModel())

        assert list_phase.variable.prior_count == 0

        list_phase.prop.one = galaxy_model

        assert list_phase.variable.prior_count == 1
        assert list_phase.variable.one == galaxy_model

        instance = list_phase.variable.instance_from_prior_medians()

        assert instance.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_constant(self, list_phase):
        galaxy = g.Galaxy()
        list_phase.prop = dict(one=galaxy)

        assert list_phase.variable.prior_count == 0
        assert list_phase.constant.one == galaxy

    def test_singular_model_info(self, list_phase):
        galaxy_model = gp.GalaxyModel(variable_redshift=True)
        list_phase.prop = dict(one=galaxy_model)

        assert len(list_phase.variable.flat_prior_model_tuples) == 1
        assert len(list_phase.variable.info.split('\n')) == 4

    def test_shared_priors(self, list_phase):
        list_phase.prop = dict(one=gp.GalaxyModel(variable_redshift=True),
                               two=gp.GalaxyModel(variable_redshift=True))

        assert list_phase.variable.prior_count == 2

        # noinspection PyUnresolvedReferences
        list_phase.prop.one.redshift = list_phase.prop.two.redshift

        assert list_phase.variable.prior_count == 1

    def test_hasattr(self, list_phase):
        list_phase.prop = dict()

        assert not hasattr(list_phase.prop, "one")
        list_phase.prop = dict(one=gp.GalaxyModel(variable_redshift=True))

        assert hasattr(list_phase.prop, "one")


def assert_ordered(items):
    assert [n for n in range(len(items))] == [item.position for item in items]



from matplotlib import pyplot as plt

from autolens.data.array.plotters import plotter_util
from autolens.data.imaging.plotters import imaging_plotters
from autolens.data.fitting.plotters import fitting_plotters


def plot_fitting_subplot(fit, should_plot_mask=True, positions=None,
                         units='arcsec', figsize=None, aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                         linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                         mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                         output_path=None, output_filename='sensitivity_fit', output_format='show'):
    """Plot the model data of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.sensitivity_fitting.SensitivityProfileFit
        Class containing fitting between the model data and observed lensing data (including residuals, chi_squareds etc.)
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=9)

    mask = fitting_plotters.get_mask(fit=fit.fit_normal, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer_normal.image_plane.kpc_per_arcsec_proper

    imaging_plotters.plot_image(image=fit.fit_normal.images[0], mask=mask, positions=positions, image_plane_pix_grid=None,
                                as_subplot=True,
                                units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                linscale=linscale,
                                cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                xyticksize=xyticksize,
                                grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                                mask_pointsize=mask_pointsize,
                                output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 4)

    fitting_plotters.plot_model_image(fit=fit.fit_normal, image_index=0, mask=mask, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 5)

    fitting_plotters.plot_residuals(fit=fit.fit_normal, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 6)

    fitting_plotters.plot_chi_squareds(fit=fit.fit_normal, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                       output_path=output_path, output_filename='', output_format=output_format)
    
    plt.subplot(rows, columns, 7)

    fitting_plotters.plot_model_image(fit=fit.fit_sensitive, image_index=0, mask=mask, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 8)

    fitting_plotters.plot_residuals(fit=fit.fit_sensitive, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 9)

    fitting_plotters.plot_chi_squareds(fit=fit.fit_sensitive, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()